"""Utilities for loading and importing solvers

.. autofunction:: available_solvers
.. autofunction:: import_module_solver
.. autofunction:: import_cls_simul

"""

import sys

if sys.version_info[:2] < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


def available_solvers(entrypoint_grp):
    """Returns a dictionary of all registered solver modules registered as an
    entrypoint_ group. Each entrypoint_ would be a key-value pair - a solver
    short name and the full import path of the solver module respectively.

    .. _entrypoint: https://packaging.python.org/guides/creating-and-discovering-plugins/#using-package-metadata

    Parameters
    ----------
    entrypoint_grp: str
        The name of the entrypoint group listing the solvers.

    """  # noqa
    return entry_points(group=entrypoint_grp)


def import_module_solver(key, entrypoint_grp):
    """Import the solver module.

    Parameters
    ----------
    key: str
        The short name of a solver.

    entrypoint_grp: str
        The name of the entrypoint group listing the solvers.

    """
    solvers = available_solvers(entrypoint_grp)
    try:
        solver = solvers[key].load()
    except KeyError:
        raise ValueError(
            "You have to give a proper solver key. Given: "
            f"{key}. Expected one of: {list(solvers)}"
        )
    else:
        return solver


def import_cls_simul(key, entrypoint_grp):
    """Import the Simul class of a solver.

    Parameters
    ----------
    key: str
        The short name of a solver.

    entrypoint_grp: str
        The name of the entrypoint group listing the solvers.

    """

    if key.startswith(entrypoint_grp + "."):
        key = key[len(entrypoint_grp) + 1 :]

    solver = import_module_solver(key, entrypoint_grp)
    return solver.Simul



from fluidsim_core.output.phys_fields_snek5000 import PhysFields4Snek5000


def test_hexa_movies(false_output):
    phys_fields = PhysFields4Snek5000(false_output)
    phys_fields.animate(dt_frame_in_sec=1e-4, normalize_vectors=True, clim=(0, 1))
    phys_fields.movies.update_animation(1)

    phys_fields.animate(
        dt_frame_in_sec=1e-4,
        normalize_vectors=True,
        clim=(0, 1),
        equation="y=0.5",
    )
    phys_fields.movies.update_animation(1)



"""Utilities for paths"""

from typing import Union
from pathlib import Path

from fluiddyn.io import FLUIDSIM_PATH

try:
    path_dir_results = Path(FLUIDSIM_PATH)
except TypeError:
    # to be able to import for transonic
    path_dir_results = None


def find_path_result_dir(thing: Union[str, Path, None] = None):
    """Return the path of a result directory.

    thing: str or Path, optional

      Can be an absolute path, a relative path, or even simply just
      the name of the directory under $FLUIDSIM_PATH.

    """
    if thing is None:
        return Path.cwd()

    if not isinstance(thing, Path):
        path = Path(thing)
    else:
        path = thing

    path = path.expanduser()

    if path.is_dir():
        return path.absolute()

    if not path.is_absolute():
        path = path_dir_results / path

    if not path.is_dir():
        raise ValueError(f"Cannot find a path corresponding to {thing}")

    return path



"""
Pure-Python core library for FluidSim framework. This package provides
generic base classes and utilities to build new solvers.

.. autosummary::
   :toctree:

   params
   info
   solver
   output
   loader
   magic
   extend_simul
   hexa_files
   scripts
   paths

"""

from ._version import __version__



"""Utilities for hexahedral files
---------------------------------

.. autofunction:: get_edges_2d

.. autoclass:: HexaField
   :members:
   :private-members:
   :undoc-members:

.. autoclass:: SetOfPhysFieldFiles
   :members:
   :private-members:
   :undoc-members:

"""

from functools import lru_cache
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt

import pymech
from pymech.neksuite.field import read_header

from fluidsim_core.output.phys_fields import SetOfPhysFieldFilesBase


def get_edges_2d(var):
    ny, nx = var.shape
    edges = np.empty([ny + 1, nx + 1])
    for iy in [0, -1]:
        for ix in [0, -1]:
            edges[iy, ix] = var[iy, ix]

    for ix in [0, -1]:
        edges[1:-1, ix] = 0.5 * (var[:-1, ix] + var[1:, ix])

    for iy in [0, -1]:
        edges[iy, 1:-1] = 0.5 * (var[iy, :-1] + var[iy, 1:])

    edges[1:-1, 1:-1] = 0.25 * (
        var[:-1, :-1] + var[1:, 1:] + var[1:, :-1] + var[:-1, 1:]
    )
    return edges


class HexaField:
    def _init_from_arrays(self, arrays):
        self.arrays = [arr for arr in arrays]

    def _init_from_hexa_data(self, hexa_data, key, equation):
        if key.startswith("v"):
            name_attr = "vel"
            if key == "vx":
                index_var = 0
            elif key == "vy":
                index_var = 1
            elif key == "vz":
                index_var = 2
            else:
                raise ValueError
        elif key in "xyz":
            name_attr = "pos"
            if key == "x":
                index_var = 0
            elif key == "y":
                index_var = 1
            elif key == "z":
                index_var = 2
            else:
                raise ValueError

        elif key.startswith("temp"):
            name_attr = "temp"
            index_var = 0
        elif key.startswith("pres"):
            name_attr = "pres"
            index_var = 0
        elif key.startswith("scalar"):
            name_attr = "scal"
            parts = key.split()
            if len(parts) == 1:
                index_var = 0
            elif len(parts) == 2:
                index_var = int(parts[1])
            else:
                raise ValueError(
                    "For scalar, key should be of the form 'scalar 2'"
                )
        else:
            raise NotImplementedError

        if equation is None:
            equation = "z=0"

        equation.replace(" ", "")

        self.arrays = []
        self.elements = []
        for elem in hexa_data.elem:
            arr_3d = getattr(elem, name_attr)[index_var]
            if equation.startswith("z="):
                z_target = float(equation[2:])
                z_3d = elem.pos[2]
                if not (z_3d.min() <= z_target <= z_3d.max()):
                    continue
                z_1d = z_3d[:, 0, 0]
                index_z = np.argmin(abs(z_1d - z_target))
                index_y = index_x = slice(None)
            elif equation.startswith("y="):
                y_target = float(equation[2:])
                y_3d = elem.pos[1]
                if not (y_3d.min() <= y_target <= y_3d.max()):
                    continue
                y_1d = y_3d[:, 0, 0]
                index_y = np.argmin(abs(y_1d - y_target))
                index_z = index_x = slice(None)
            elif equation.startswith("x="):
                x_target = float(equation[2:])
                x_3d = elem.pos[0]
                if not (x_3d.min() <= x_target <= x_3d.max()):
                    continue
                x_1d = x_3d[:, 0, 0]
                index_x = np.argmin(abs(x_1d - x_target))
                index_z = index_y = slice(None)
            else:
                raise NotImplementedError

            arr = arr_3d[index_z, index_y, index_x]

            self.arrays.append(arr)

            dict_elem = {"array": arr}

            if key in "xyz":
                dict_elem["edges"] = get_edges_2d(arr)
                dict_elem["lims"] = arr.min(), arr.max()

            self.elements.append(dict_elem)

        if key in "xyz":
            self.lims = hexa_data.lims.pos[index_var]

        self.time = hexa_data.time

    def __init__(
        self, key, hexa_data=None, arrays=None, time=None, equation="z=0"
    ):
        self.key = key
        self.equation = equation
        if hexa_data is None and arrays is not None:
            self._init_from_arrays(arrays)
        elif hexa_data is not None and arrays is None:
            self._init_from_hexa_data(hexa_data, key, equation)
        else:
            raise ValueError

        if time is not None:
            self.time = time

    def __mul__(self, arg):
        return self.__class__(
            self.key,
            arrays=[arg * arr for arr in self.arrays],
            time=self.time,
            equation=self.equation,
        )

    __rmul__ = __mul__

    def __add__(self, arg):
        return self.__class__(
            self.key,
            arrays=[arr0 + arr1 for arr0, arr1 in zip(arg.arrays, self.arrays)],
            time=(self.time + arg.time) / 2,
            equation=self.equation,
        )

    def min(self):
        result = np.inf
        for arr in self.arrays:
            min_elem = arr.min()
            if min_elem < result:
                result = min_elem
        return result

    def max(self):
        result = -np.inf
        for arr in self.arrays:
            max_elem = arr.max()
            if max_elem > result:
                result = max_elem
        return result


class SetOfPhysFieldFiles(SetOfPhysFieldFilesBase):
    def __init__(self, path_dir=None, output=None, prefix=None):
        self.prefix = prefix
        self._other_sets_of_files = {}
        super().__init__(path_dir=path_dir, output=output)

    def get_dataset_from_time(self, time):
        index = self.times.tolist().index(time)
        return self.get_dataset_from_path(self.path_files[index])

    @lru_cache(maxsize=2)
    def get_dataset_from_path(self, path):
        return pymech.open_dataset(path)

    def read_hexadata_from_time(self, time, skip_vars=(), prefix=None):
        sof = self._get_setoffiles_from_prefix(prefix)
        try:
            index = sof.times.tolist().index(time)
        except ValueError:
            print(f"available times: {self.times}")
            raise
        return sof._read_hexadata_from_path(
            sof.path_files[index], skip_vars=skip_vars
        )

    def read_hexadata(self, path=None, index=None, skip_vars=(), prefix=None):
        if index is not None and path is not None:
            raise ValueError("path and index are both not None")
        elif index is None and path is None:
            index = -1

        sof = self._get_setoffiles_from_prefix(prefix)

        if index is not None:
            path = sof.path_files[index]

        if not Path(path).exists():
            raise ValueError(
                f"{path = } does not exists. Available path: {sof.path_files}"
            )

        return sof._read_hexadata_from_path(path, skip_vars=skip_vars)

    @lru_cache(maxsize=2)
    def _read_hexadata_from_path(self, path, skip_vars=()):
        return pymech.readnek(path, skip_vars=skip_vars)

    def _get_field_to_plot_from_file(
        self, path_file, key, equation, skip_vars=()
    ):
        hexa_data = self._read_hexadata_from_path(path_file, skip_vars=skip_vars)
        hexa_field = HexaField(key, hexa_data, equation=equation)
        return hexa_field, float(hexa_data.time)

    def init_hexa_pcolormesh(
        self, ax, hexa_color, hexa_x, hexa_y, vmin=None, vmax=None, **kwargs
    ):
        images = []

        if vmax is None:
            vmax = hexa_color.max()

        if vmin is None:
            vmin = hexa_color.min()

        for arr, elem_x, elem_y in zip(
            hexa_color.arrays, hexa_x.elements, hexa_y.elements
        ):
            x_edges = elem_x["edges"]
            y_edges = elem_y["edges"]

            image = ax.pcolormesh(
                x_edges,
                y_edges,
                arr,
                shading="flat",
                vmin=vmin,
                vmax=vmax,
                **kwargs,
            )

            images.append(image)

        fig = ax.figure

        cbar = fig.colorbar(images[0])

        def on_move(event):
            if event.inaxes is not None and event.inaxes == ax:
                x = event.xdata
                y = event.ydata

                elements_possibly_touched = []

                for image, elem_x, elem_y in zip(
                    images, hexa_x.elements, hexa_y.elements
                ):
                    xmin, xmax = elem_x["lims"]
                    ymin, ymax = elem_y["lims"]

                    x_2d = elem_x["array"]
                    y_2d = elem_y["array"]

                    if (xmin <= x <= xmax) and (ymin <= y <= ymax):
                        distance2_2d = (x_2d - x) ** 2 + (y_2d - y) ** 2
                        i1d = distance2_2d.argmin()
                        iy, ix = np.unravel_index(i1d, distance2_2d.shape)
                        distance2_min = distance2_2d[iy, ix]
                        arr = image.get_array()
                        if arr.ndim > 1:
                            # matplotlib >= 3.8.0
                            arr = arr.flatten()
                        color = arr[i1d]
                        elements_possibly_touched.append((distance2_min, color))

                if elements_possibly_touched:
                    elements_possibly_touched = sorted(
                        elements_possibly_touched, key=lambda el: el[0]
                    )
                    element_touched = elements_possibly_touched[0]
                    _, color = element_touched

                    message = (
                        ax.format_coord(x, y) + f" {hexa_color.key} = {color:.3f}"
                    )

                    try:
                        set_message = fig.canvas.toolbar.set_message
                    except AttributeError:
                        # for testing and coverage
                        pass
                    else:
                        set_message(message)

        fig.canvas.mpl_connect("motion_notify_event", on_move)
        # to be able to test this callback
        fig._on_move_hexa = on_move

        return images, cbar

    def init_quiver_1st_step(self, hexa_x, hexa_y, percentage_dx_quiver=4.0):
        xmin = hexa_x.min()
        xmax = hexa_x.max()

        ymin = hexa_y.min()
        ymax = hexa_y.max()

        dx_quiver = percentage_dx_quiver / 100 * (xmax - xmin)
        nx_quiver = int((xmax - xmin) / dx_quiver)
        ny_quiver = int((ymax - ymin) / dx_quiver)

        x_approx_quiver = np.linspace(
            xmin + dx_quiver, xmax - dx_quiver, nx_quiver
        )
        y_approx_quiver = np.linspace(
            xmin + dx_quiver, ymax - dx_quiver, ny_quiver
        )

        indices_vectors_in_elems = []

        for x_2d, y_2d in zip(hexa_x.arrays, hexa_y.arrays):
            xmin = x_2d.min()
            xmax = x_2d.max()
            ymin = y_2d.min()
            ymax = y_2d.max()

            indices_vectors_in_elem = []

            for y_approx in y_approx_quiver:
                if y_approx < ymin:
                    continue
                if y_approx > ymax:
                    break
                for x_approx in x_approx_quiver:
                    if x_approx < xmin:
                        continue
                    if x_approx > xmax:
                        break

                    distance2_2d = (x_2d - x_approx) ** 2 + (y_2d - y_approx) ** 2

                    iy, ix = np.unravel_index(
                        distance2_2d.argmin(), distance2_2d.shape
                    )
                    indices_vectors_in_elem.append((iy, ix))

            indices_vectors_in_elems.append(sorted(set(indices_vectors_in_elem)))

        x_quiver = []
        y_quiver = []

        for indices_vectors_in_elem, x_2d, y_2d in zip(
            indices_vectors_in_elems, hexa_x.arrays, hexa_y.arrays
        ):
            for iy, ix in indices_vectors_in_elem:
                x_quiver.append(x_2d[iy, ix])
                y_quiver.append(y_2d[iy, ix])

        return indices_vectors_in_elems, x_quiver, y_quiver

    def compute_vectors_quiver(self, hexa_vx, hexa_vy, indices_vectors_in_elems):
        vx_quiver = []
        vy_quiver = []
        vmax = -np.inf

        for indices_vectors_in_elem, arr_vx, arr_vy in zip(
            indices_vectors_in_elems, hexa_vx.arrays, hexa_vy.arrays
        ):
            vmax_elem = np.max(np.sqrt(arr_vx**2 + arr_vy**2))
            if vmax_elem > vmax:
                vmax = vmax_elem

            for iy, ix in indices_vectors_in_elem:
                vx_quiver.append(arr_vx[iy, ix])
                vy_quiver.append(arr_vy[iy, ix])

        return np.array(vx_quiver), np.array(vy_quiver), vmax

    @lru_cache
    def get_letters_axes_from_equation(self, equation):
        equation = equation.replace(" ", "")
        if equation.startswith("z="):
            letter_x_axis = "x"
            letter_y_axis = "y"
        elif equation.startswith("y="):
            letter_x_axis = "x"
            letter_y_axis = "z"
        elif equation.startswith("x="):
            letter_x_axis = "z"
            letter_y_axis = "y"
        else:
            raise NotImplementedError
        return letter_x_axis, letter_y_axis

    @lru_cache
    def _get_setoffiles_from_prefix(self, prefix):
        if prefix is None:
            return self
        else:
            return type(self)(
                path_dir=self.path_dir, output=self.output, prefix=prefix
            )

    def plot_hexa(
        self,
        key=None,
        time=None,
        equation="z=0",
        percentage_dx_quiver=4.0,
        vmin=None,
        vmax=None,
        normalize_vectors=True,
        quiver_kw={},
        pcolor_kw={},
        prefix=None,
    ):
        sof_color = self._get_setoffiles_from_prefix(prefix)
        times_color = sof_color.times

        if time is None:
            time = times_color[-1]
        else:
            time = times_color[abs(times_color - time).argmin()]

        if time not in self.times:
            raise ValueError

        hexa_data = self.read_hexadata_from_time(time)
        if prefix is None:
            hexa_data_color = hexa_data
        else:
            hexa_data_color = sof_color.read_hexadata_from_time(time)

        key_field = sof_color.get_key_field_to_plot(key)
        hexa_field = HexaField(key_field, hexa_data_color, equation=equation)

        letter_x_axis, letter_y_axis = self.get_letters_axes_from_equation(
            equation
        )
        hexa_x = HexaField(letter_x_axis, hexa_data, equation=equation)
        hexa_y = HexaField(letter_y_axis, hexa_data, equation=equation)
        hexa_vx = HexaField("v" + letter_x_axis, hexa_data, equation=equation)
        hexa_vy = HexaField("v" + letter_y_axis, hexa_data, equation=equation)

        fig, ax = plt.subplots(layout="constrained")

        self.init_hexa_pcolormesh(
            ax, hexa_field, hexa_x, hexa_y, vmin=vmin, vmax=vmax, **pcolor_kw
        )

        indices_vectors_in_elems, x_quiver, y_quiver = self.init_quiver_1st_step(
            hexa_x, hexa_y, percentage_dx_quiver=percentage_dx_quiver
        )
        vx_quiver, vy_quiver, vmax = self.compute_vectors_quiver(
            hexa_vx, hexa_vy, indices_vectors_in_elems
        )

        if normalize_vectors:
            vx_quiver /= vmax
            vy_quiver /= vmax

        ax.quiver(x_quiver, y_quiver, vx_quiver, vy_quiver, **quiver_kw)

        ax.set_xlabel(f"${letter_x_axis}$")
        ax.set_ylabel(f"${letter_y_axis}$")
        title = f"{key_field}, $t = {time:.3f}$"
        if vmax is not None:
            title += r", $|\vec{v}|_{max} = $" + f"{vmax:.3f}"
        ax.set_title(title)

    def time_from_path(self, path):
        header = self.get_header(path)
        return header.time

    def get_header(self, path=None):
        if path is None:
            path = self.path_files[0]
        return read_header(path)

    def _get_glob_pattern(self):
        prefix = "" if self.prefix is None else self.prefix
        session_id = self.output.sim.params.output.session_id
        case = self.output.name_solver
        return f"session_{session_id:02d}/{prefix}{case}0.f?????"

    def get_vector_for_plot(self, time, equation=None, skip_vars=()):
        if equation is None:
            equation = "z=0"
        letter_x_axis, letter_y_axis = self.get_letters_axes_from_equation(
            equation
        )
        # temporary hack
        time = self.times[abs(self.times - time).argmin()]
        hexa_data = self.read_hexadata_from_time(time, skip_vars=skip_vars)
        vec_xaxis = HexaField(
            hexa_data=hexa_data, key="v" + letter_x_axis, equation=equation
        )
        vec_yaxis = HexaField(
            hexa_data=hexa_data, key="v" + letter_y_axis, equation=equation
        )
        return vec_xaxis, vec_yaxis

    def get_key_field_to_plot(self, key_prefered=None):
        if key_prefered is None:
            header = self.get_header()
            if "T" in header.variables:
                return "temperature"
            elif "P" in header.variables:
                return "pressure"
            else:
                return "scalar"
        else:
            return key_prefered



from dataclasses import dataclass

import matplotlib.pyplot as plt

import pymech

from fluidsim_core.hexa_files import SetOfPhysFieldFiles, HexaField


@dataclass
class MockEvent:
    inaxes: object
    xdata: float
    ydata: float


def test_setoffiles(false_output):
    output = false_output
    assert output.path_run.exists()

    hexa_data_loaded = pymech.readnek(
        output.path_run / f"session_00/{output.name_solver}0.f{1:05d}"
    )

    HexaField(hexa_data=hexa_data_loaded, key="vz")
    HexaField(hexa_data=hexa_data_loaded, key="pres", equation=None)

    HexaField(hexa_data=hexa_data_loaded, key="scalar", equation=None)
    HexaField(hexa_data=hexa_data_loaded, key="scalar 1", equation=None)

    set_of_files = SetOfPhysFieldFiles(output.path_run, output=output)
    hexa_z, time = set_of_files.get_field_to_plot(2, key="z", equation=None)

    set_of_files.get_key_field_to_plot("pressure")

    2.0 * hexa_z + hexa_z

    set_of_files.plot_hexa()
    set_of_files.plot_hexa(equation="y=0.75")
    set_of_files.plot_hexa(equation="x=0.5")

    set_of_files.plot_hexa(key="scalar 1", prefix="sts")
    set_of_files.plot_hexa(prefix="sts", time=0)

    ax = plt.gca()
    on_move = ax.figure._on_move_hexa
    event = MockEvent(ax, 0.5, 0.5)
    on_move(event)

    set_of_files.get_vector_for_plot(time=3)
    set_of_files.get_dataset_from_time(2.5)

    set_of_files.read_hexadata()
    set_of_files.read_hexadata(prefix="sts")



from fluidsim_core.info import InfoSolverCore
from fluidsim_core.solver import SimulCore


class InfoSolverTest(InfoSolverCore):
    """Contain the information on a solver."""

    def _init_root(self):
        super()._init_root()

        self._set_attribs(
            {
                "module_name": "fluidsim_core.tests.test_solver",
                "class_name": "SimulTest",
                "short_name": "Test",
            }
        )


class SimulTest(SimulCore):
    InfoSolver = InfoSolverTest

    @staticmethod
    def _complete_params_with_default(params):
        params._set_attribs({"foo": 42, "bar": True})

    @classmethod
    def create_default_params(cls):
        return super().create_default_params()

    def __init__(self, params):
        super().__init__(params)


Simul = SimulTest



"""Movies output
================

Contains base classes which acts as a framework to
implement the method ``animate`` to make movies.

Provides:

.. autoclass:: MoviesBase
   :members:
   :private-members:
   :undoc-members:

.. autoclass:: MoviesBase1D
   :members:
   :private-members:
   :undoc-members:

.. autoclass:: MoviesBase2D
   :members:
   :private-members:
   :undoc-members:

.. autoclass:: MoviesBasePhysFields
   :members:
   :private-members:
   :undoc-members:

.. autoclass:: MoviesBasePhysFieldsHexa
   :members:
   :private-members:
   :undoc-members:

"""

from abc import abstractmethod
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
from matplotlib.widgets import Button
import mpl_toolkits.axes_grid1

from fluiddyn.util import mpi, is_run_from_jupyter


class MoviesBase:
    """Base class defining most generic functions for movies."""

    def __init__(self, output):
        params = output.sim.params
        self.output = output
        self.sim = output.sim
        self.params = params.output
        self.oper = self.sim.oper

        self._set_font()

    def _set_font(self, family="serif", size=12):
        """Use to set font attribute. May be either an alias (generic name
        is CSS parlance), such as serif, sans-serif, cursive, fantasy, or
        monospace, a real font name or a list of real font names.

        """
        self.font = {
            "family": family,
            "color": "black",
            "weight": "normal",
            "size": size,
        }

    def init_animation(
        self, key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
    ):
        """Initializes animated fig. and list of times of save files to load."""
        self._set_key_field(key_field)
        self._init_ani_times(tmin, tmax, dt_equations)
        self.fig, self.ax = plt.subplots(
            num=numfig, layout="constrained", **fig_kw
        )
        self._init_labels()

        if not self._interactive:
            return

        # see https://stackoverflow.com/a/44989063
        playerax = self.fig.add_axes([0.05, 0.015, 0.22, 0.04])
        divider = mpl_toolkits.axes_grid1.make_axes_locatable(playerax)
        bax = divider.append_axes("right", size="80%", pad=0.05)
        sax = divider.append_axes("right", size="80%", pad=0.05)
        fax = divider.append_axes("right", size="80%", pad=0.05)
        ofax = divider.append_axes("right", size="100%", pad=0.05)

        self._buttons = []

        def init_button(ax, label, method):
            button = Button(ax, label=label)
            button.on_clicked(method)
            self._buttons.append(button)

        init_button(playerax, "$\u29CF$", self._one_backward)
        init_button(bax, "$\u25C0$", self._backward)
        init_button(sax, "$\u25A0$", self.pause)
        init_button(fax, "$\u25B6$", self._forward)
        init_button(ofax, "$\u29D0$", self._one_forward)

    def resume(self):
        self.paused = False
        self._animation.resume()

    def pause(self, event=None):
        self.paused = True
        self._animation.pause()

    def _forward(self, event=None):
        self._forwards = True
        self.resume()

    def _backward(self, event=None):
        self._forwards = False
        self.resume()

    def _one_forward(self, event=None):
        self._forwards = True
        self.onestep()

    def _one_backward(self, event=None):
        self._forwards = False
        self.onestep()

    def _get_default_tmax(self):
        try:
            return self.sim.time_stepping.t
        except AttributeError:
            return self.phys_fields.set_of_phys_files.get_max_time()

    def _init_ani_times(self, tmin, tmax, dt_equations):
        """Initialization of the variable ani_times for one animation."""
        if tmax is None or tmin is None:
            self.phys_fields.set_of_phys_files.update_times()
        if tmax is None:
            tmax = self._get_default_tmax()

        if tmin is None:
            tmin = self.phys_fields.set_of_phys_files.get_min_time()

        if tmin > tmax:
            raise ValueError(
                "Error tmin > tmax. Value tmin should be smaller than tmax"
            )

        if dt_equations is None:
            dt_equations = self.params.periods_save.phys_fields

        if dt_equations > tmax:
            self.ani_times = np.array([tmin, tmax])
        else:
            self.ani_times = np.arange(tmin, tmax, dt_equations)

    @abstractmethod
    def update_animation(self, frame, **fargs):
        """Replace this function to load data for next frame and update the
        figure.

        """

    def get_field_to_plot(self, time=None, key=None, equation=None):
        """
        Once a saved file is loaded, this selects the field and mpi-gathers.

        Returns
        -------
        field : nd array or string
        """
        raise NotImplementedError(
            "get_field_to_plot function declaration missing."
        )

    def _init_labels(self, xlabel=None, ylabel=None):
        """Initialize the labels."""
        if xlabel is None:
            xlabel = self.sim.oper.axes[1]
        if ylabel is None:
            ylabel = self.sim.oper.axes[0]
        if xlabel:
            self.ax.set_xlabel(xlabel, fontdict=self.font)
        if ylabel:
            self.ax.set_ylabel(ylabel, fontdict=self.font)

    def _get_axis_data(self):
        """Replace this function to load axis data."""
        raise NotImplementedError("_get_axis_data  function declaration missing.")

    def _set_key_field(self, key_field=None):
        """
        Defines key_field default.
        """
        self.key_field = self.phys_fields.get_key_field_to_plot(
            forbid_compute=True, key_prefered=key_field
        )

    def animate(
        self,
        key_field=None,
        dt_frame_in_sec=0.3,
        dt_equations=None,
        tmin=None,
        tmax=None,
        repeat=True,
        save_file=False,
        numfig=None,
        interactive=None,
        fargs={},
        fig_kw={},
        **kwargs,
    ):
        """Load the key field from multiple save files and display as
        an animated plot or save as a movie file.

        Parameters
        ----------
        key_field : str
            Specifies which field to animate
        dt_frame_in_sec : float
            Interval between animated frames in seconds
        dt_equations : float
            Approx. interval between saved files to load in simulation time
            units
        tmax : float
            Animate till time `tmax`.
        repeat : bool
            Loop the animation
        save_file : str or bool
            Path to save the movie. When `True`  saves into a file instead
            of plotting it on screen (default: ~/fluidsim_movie.mp4). Specify
            a string to save to another file location. Format is autodetected
            from the filename extension.
        numfig : int
            Figure number on the window
        interactive : bool
            Add player buttons (pause, step by step and forward/backward)
        fargs : dict
            Dictionary of arguments for `update_animation`. Matplotlib
            requirement.
        fig_kw : dict
            Dictionary for arguments for the figure.

        Other Parameters
        ----------------
        All `kwargs` are passed on to `init_animation` and `_ani_save`

        xmax : float
            Set x-axis limit for 1D animated plots
        ymax : float
            Set y-axis limit for 1D animated plots
        clim : tuple
            Set colorbar limits for 2D animated plots
        step : int
            Set step value to get a coarse 2D field
        QUIVER : bool
            Set quiver on or off on top of 2D pcolor plots
        normalize_vectors : bool (default True)
            Normalize the vectors at each time by the instantaneous maximum
        quiver_kw: dict
            Dictionary for arguments for the quiver call.
        pcolor_kw: dict
            Dictionary for arguments for the pcolormesh call.
        codec : str
            Codec used to save into a movie file (default: ffmpeg)

        Examples
        --------
        >>> import fluidsim as fls
        >>> sim = fls.load_sim_for_plot()
        >>> animate = sim.output.spectra.animate
        >>> animate('E')
        >>> animate('rot')
        >>> animate('rot', dt_equations=0.1, dt_frame_in_sec=50, clim=(-5, 5))
        >>> animate('rot', clim=(-300, 300), fig_kw={"figsize": (14, 4)})
        >>> animate('rot', tmax=25, clim=(-5, 5), save_file='True')
        >>> animate('rot', clim=(-5, 5), save_file='~/fluidsim.gif', codec='imagemagick')

        .. TODO: Use FuncAnimation with blit=True option.

        Notes
        -----

        This method is kept as generic as possible. Any arguments specific to
        1D, 2D or 3D animations or specific to a type of output are be passed
        via keyword arguments (``kwargs``) into its respective
        ``init_animation`` or ``update_animation`` methods.

        """
        if mpi.rank > 0:
            raise NotImplementedError("Do NOT use this function with MPI !")

        if interactive and save_file:
            raise ValueError("Incompatible options interactive and save_file")

        self._interactive = interactive
        self.init_animation(
            key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
        )

        if isinstance(repeat, int) and repeat:
            nb_repeat = repeat
            repeat = False
        else:
            nb_repeat = 1

        self._min = 0
        self._max = nb_repeat * len(self.ani_times) - 1
        frames = self._max + 1
        if interactive:
            frames = self._frames_iterative
            self._forwards = True
            self._index = self._min = 0

        self._animation = animation.FuncAnimation(
            self.fig,
            self.update_animation,
            frames=frames,
            fargs=fargs.items(),
            interval=dt_frame_in_sec * 1000,
            blit=False,
            repeat=repeat,
            save_count=self._max,
        )

        self.paused = False

        if save_file:
            if not isinstance(save_file, str):
                save_file = r"~/fluidsim_movie.mp4"

            self._ani_save(save_file, dt_frame_in_sec, **kwargs)
            return

        self.fig.canvas.mpl_connect("button_press_event", self._toggle_pause)

    def _frames_iterative(self):
        while not self.paused:
            self._index += self._forwards - (not self._forwards)
            frame = self._index
            if frame < self._min:
                self._index = frame = self._max
            elif frame > self._max:
                self._index = frame = self._min
            elif frame == self._max or frame == self._min:
                self.pause()
            yield frame

    def onestep(self):
        self.pause()
        if self._index > self._min and self._index < self._max:
            self._index += self._forwards - (not self._forwards)
        elif self._index == self._min:
            if self._forwards:
                self._index += 1
            else:
                self._index = self._max
        elif self._index == self._max:
            if not self._forwards:
                self._index -= 1
            else:
                self._index = self._min

        self.update_animation(self._index)
        self.fig.canvas.draw_idle()

    def _toggle_pause(self, event):
        if event.inaxes != self.fig.axes[0]:
            return
        if self.paused:
            self._animation.resume()
        else:
            self._animation.pause()
        self.paused = not self.paused

    def interact(
        self,
        key_field=None,
        dt_equations=None,
        tmin=None,
        tmax=None,
        fig_kw={},
        **kwargs,
    ):
        """Launches an interactive plot.

        Parameters
        ----------
        key_field : str
            Specifies which field to animate
        dt_equations : float
            Approx. interval between saved files to load in simulation time
            units
        tmax : float
            Animate till time `tmax`.
        fig_kw : dict
            Dictionary for arguments for the figure.

        Other Parameters
        ----------------
        All `kwargs` are passed on to `init_animation` and `_ani_save`

        xmax : float
            Set x-axis limit for 1D animated plots
        ymax : float
            Set y-axis limit for 1D animated plots
        clim : tuple
            Set colorbar limits for 2D animated plots
        step : int
            Set step value to get a coarse 2D field
        QUIVER : bool
            Set quiver on or off on top of 2D pcolor plots
        quiver_kw: dict
            Dictionary for arguments for the quiver call.

        Notes
        -----
        Installation instructions for notebook::

          pip install ipywidgets
          jupyter nbextension enable --py widgetsnbextension

        Restart the notebook and call the function using::

          >>> %matplotlib notebook

        For JupyterLab::

          pip install ipywidgets ipympl
          jupyter labextension install @jupyter-widgets/jupyterlab-manager

        Restart JupyterLab and call the function using::

          >>> %matplotlib widget

        """
        try:
            from ipywidgets import interact, widgets
        except ImportError as exc:
            raise ImportError(
                "See fluidsim_core.output.movies.interact docstring."
            ) from exc

        if not is_run_from_jupyter():
            raise ValueError("Works only inside Jupyter.")

        self._interactive = False
        self.init_animation(
            key_field, 0, dt_equations, tmin, tmax, fig_kw, **kwargs
        )
        if tmin is None:
            tmin = self.ani_times[0]
        if tmax is None:
            tmax = self.ani_times[-1]
        if dt_equations is None:
            dt_equations = self.ani_times[1] - self.ani_times[0]

        slider = widgets.FloatSlider(
            min=float(tmin),
            max=float(tmax),
            step=float(dt_equations),
            value=float(tmin),
        )

        def widget_update(time):
            frame = np.argmin(abs(self.ani_times - time))
            self.update_animation(frame)
            self.fig.canvas.draw()

        interact(widget_update, time=slider)

    def _ani_save(self, path_file, dt_frame_in_sec, codec="ffmpeg", **kwargs):
        """Saves the animation using `matplotlib.animation.writers`."""

        path_file = os.path.expandvars(path_file)
        path_file = os.path.expanduser(path_file)
        avail_writers = animation.writers.list()
        if len(avail_writers) == 0:
            raise ValueError(
                "Please install a codec library. For e.g. ffmpeg, mencoder, "
                "imagemagick, html"
            )

        elif codec not in avail_writers:
            print(f"Using one of the available codecs: {avail_writers}")
            codec = avail_writers[0]

        Writer = animation.writers[codec]

        print("Saving movie to ", path_file, "...")
        writer = Writer(
            fps=1.0 / dt_frame_in_sec, metadata=dict(artist="FluidSim")
        )
        # _animation is a FuncAnimation object
        self._animation.save(path_file, writer=writer, dpi=150)


class MoviesBase1D(MoviesBase):
    """Base class defining most generic functions for movies for 1D data."""

    def _init_labels(self, xlabel=None, ylabel=None):
        """Initialize the labels."""
        if xlabel is None:
            xlabel = self.sim.oper.axes[0]
        if ylabel is None:
            ylabel = False
        super()._init_labels(xlabel, ylabel)

    def init_animation(
        self, key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
    ):
        """Initializes animated figure."""

        super().init_animation(
            key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
        )

        ax = self.ax
        (self._ani_line,) = ax.plot([], [])

        if "xmax" in kwargs:
            ax.set_xlim(0, kwargs["xmax"])
        else:
            ax.set_xlim(0, self.output.sim.oper.lx)

        if "ymax" in kwargs:
            ax.set_ylim(1e-16, kwargs["ymax"])

    def get_field_to_plot(self, time, key):
        return self.phys_fields.get_field_to_plot(time=time, key=self.key_field)

    def update_animation(self, frame, **fargs):
        """Loads contour data and updates figure."""
        time = self.ani_times[frame % len(self.ani_times)]
        y, time = self.get_field_to_plot(time=time, key=self.key_field)
        x = self._get_axis_data()

        self._ani_line.set_data(x, y)
        self.ax.set_title(
            self.key_field + f", $t = {time:.3f}$\n" + self.output.summary_simul
        )

        return self._ani_line

    def _get_axis_data(self):
        """Get axis data.

        Returns
        -------

        x : array
          x-axis data.

        """
        try:
            x = self.oper.xs
        except AttributeError:
            x = self.oper.x_seq
        return x


class MoviesBase2D(MoviesBase):
    """Base class defining most generic functions for movies for 2D data."""

    def _get_axis_data(self):
        """Get axis data.

        Returns
        -------

        x : array
          x-axis data.

        y : array
          y-axis data.

        """
        return self.phys_fields._get_axis_data(self.phys_fields._equation)


class MoviesBasePhysFields(MoviesBase2D):
    def __init__(self, output, phys_fields):
        self.phys_fields = phys_fields
        super().__init__(output)

    def init_animation(
        self, key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
    ):
        """Initialize list of files and times, pcolor plot, quiver and colorbar."""
        self.phys_fields.set_of_phys_files.update_times()
        self.time_files = self.phys_fields.set_of_phys_files.times

        if "equation" in kwargs:
            equation = kwargs["equation"]
            print(f"phys_fields.set_equation_crosssection({equation})")
            self.phys_fields.set_equation_crosssection(equation)

        dt_equations_reasonable = np.median(np.diff(self.time_files)) / 2
        if dt_equations is None:
            dt_equations = dt_equations_reasonable
            print(f"{dt_equations = :.4f}")

        if tmax is None:
            tmax = self.time_files.max()
        else:
            if tmax > self.time_files.max():
                raise ValueError(
                    f"{tmax=} has to be smaller or equal to "
                    f"{self.time_files.max()}."
                )

        if tmin is None:
            tmin = self.time_files.min()

        super().init_animation(
            key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
        )

        dt_file = (self.time_files[-1] - self.time_files[0]) / len(
            self.time_files
        )
        if dt_equations < dt_file / 4:
            raise ValueError(f"dt_equations < {dt_file / 4 = }")

        if dt_equations > (tmax - tmin) / 2:
            raise ValueError(
                f"{dt_equations = } > {(tmax - tmin) / 2 = }. "
                "Set dt_equations to None to get a reasonable value "
                f"({dt_equations_reasonable})"
            )

        field, time = self.phys_fields.get_field_to_plot(
            self.key_field, time=tmin
        )

        try:
            vec_xaxis, vec_yaxis = self.phys_fields.get_vector_for_plot(time=time)
        except ValueError:
            self.phys_fields._can_plot_quiver = False
            vec_xaxis = vec_yaxis = None
        else:
            self.phys_fields._can_plot_quiver = True

        self._init_fig(field, time, vec_xaxis, vec_yaxis, **kwargs)

    def _init_fig(self, field, time, vec_xaxis=None, vec_yaxis=None, **kwargs):
        """Initialize only the figure and related matplotlib objects. This
        method is shared by both ``animate`` and ``online_plot``
        functionalities.

        """
        self._step = step = 1 if "step" not in kwargs else kwargs["step"]
        self._QUIVER = True if "QUIVER" not in kwargs else kwargs["QUIVER"]
        quiver_kw = {} if "quiver_kw" not in kwargs else kwargs["quiver_kw"]
        pcolor_kw = {} if "pcolor_kw" not in kwargs else kwargs["pcolor_kw"]
        tmp_key = "normalize_vectors"
        normalize_vectors = {} if tmp_key not in kwargs else kwargs[tmp_key]

        x, y = self._get_axis_data()
        x, y = x[::step], y[::step]
        XX, YY = np.meshgrid(x, y)
        field = field[::step, ::step]
        assert (len(y), len(x)) == field.shape

        self._im = self.ax.pcolormesh(
            XX, YY, field, shading="nearest", **pcolor_kw
        )
        self._ani_cbar = self.fig.colorbar(self._im)

        vmax = None
        if self.phys_fields._can_plot_quiver and self._QUIVER:
            self._ani_quiver, vmax = self.phys_fields._quiver_plot(
                self.ax,
                vec_xaxis,
                vec_yaxis,
                XX,
                YY,
                normalize_vectors=normalize_vectors,
                skip=self.phys_fields._skip_quiver,
                **quiver_kw,
            )

        self._clim = kwargs.get("clim")
        self._set_clim()

        if not hasattr(self, "key_field"):
            self._set_key_field()

        self.phys_fields._set_title(self.ax, self.key_field, time, vmax)

    def update_animation(self, frame, **fargs):
        """Loads data and updates figure."""
        time = self.ani_times[frame % len(self.ani_times)]
        step = self._step

        field, time = self.phys_fields.get_field_to_plot(
            time=time,
            key=self.key_field,
            interpolate_time=True,
        )

        field = field[::step, ::step]

        # Update figure, quiver and colorbar
        self._im.set_array(field.flatten())
        if self.phys_fields._can_plot_quiver and self._QUIVER:
            vec_xaxis, vec_yaxis = self.phys_fields.get_vector_for_plot(time=time)

            vmax = np.max(np.sqrt(vec_xaxis**2 + vec_yaxis**2))
            skip = self.phys_fields._skip_quiver
            self._ani_quiver.set_UVC(
                vec_xaxis[::skip, ::skip] / vmax, vec_yaxis[::skip, ::skip] / vmax
            )
        else:
            vmax = None

        self._im.autoscale()
        self._set_clim()

        self.phys_fields._set_title(self.ax, self.key_field, time, vmax)

    def _set_clim(self):
        """Maintains a constant colorbar throughout the animation."""
        clim = self._clim
        if clim is not None:
            self._im.set_clim(*clim)
            ticks = np.linspace(*clim, num=11, endpoint=True)
            self._ani_cbar.set_ticks(ticks)


class MoviesBasePhysFieldsHexa(MoviesBasePhysFields):
    def _init_fig(self, field, time, vec_xaxis=None, vec_yaxis=None, **kwargs):
        """Initialize only the figure and related matplotlib objects. This
        method is shared by both ``animate`` and ``online_plot``
        functionalities.

        """
        self._step = step = 1 if "step" not in kwargs else kwargs["step"]
        self._QUIVER = True if "QUIVER" not in kwargs else kwargs["QUIVER"]
        vmin = None if "vmin" not in kwargs else kwargs["vmin"]
        vmax = None if "vmax" not in kwargs else kwargs["vmax"]
        quiver_kw = {} if "quiver_kw" not in kwargs else kwargs["quiver_kw"]
        pcolor_kw = {} if "pcolor_kw" not in kwargs else kwargs["pcolor_kw"]
        tmp_key = "normalize_vectors"
        self.normalize_vectors = {} if tmp_key not in kwargs else kwargs[tmp_key]

        if step != 1:
            raise NotImplementedError

        hexa_x, hexa_y = self._get_axis_data()
        hexa_field = field

        hexa_vec_xaxis = vec_xaxis
        hexa_vec_yaxis = vec_yaxis

        set_of_phys_files = self.phys_fields.set_of_phys_files
        ax = self.ax

        self._images, self._ani_cbar = set_of_phys_files.init_hexa_pcolormesh(
            ax, hexa_field, hexa_x, hexa_y, vmin=vmin, vmax=vmax, **pcolor_kw
        )

        self._clim = kwargs.get("clim")
        self._set_clim()

        skip_vars = []
        if self.key_field != "temperature":
            skip_vars.append("temperature")
        if self.key_field != "pressure":
            skip_vars.append("pressure")
        self._skip_vars = tuple(skip_vars)

        if self._QUIVER:
            assert hexa_x.equation == hexa_vec_xaxis.equation, (
                hexa_x.equation,
                hexa_vec_xaxis.equation,
            )
            (
                self._indices_vectors_in_elems,
                x_quiver,
                y_quiver,
            ) = set_of_phys_files.init_quiver_1st_step(
                hexa_x, hexa_y, percentage_dx_quiver=4.0
            )

            vx_quiver, vy_quiver, vmax = set_of_phys_files.compute_vectors_quiver(
                hexa_vec_xaxis, hexa_vec_yaxis, self._indices_vectors_in_elems
            )

            if self.normalize_vectors:
                vx_quiver /= vmax
                vy_quiver /= vmax

            self._ani_quiver = ax.quiver(
                x_quiver, y_quiver, vx_quiver, vy_quiver, **quiver_kw
            )
        else:
            vmax = None

        self.phys_fields._set_title(ax, self.key_field, time, vmax)

    def update_animation(self, frame, **fargs):
        """Loads data and updates figure."""
        time = self.ani_times[frame % len(self.ani_times)]
        # step = self._step

        phys_fields = self.phys_fields
        set_of_phys_files = phys_fields.set_of_phys_files

        hexa_field, time = phys_fields.get_field_to_plot(
            time=time,
            key=self.key_field,
            interpolate_time=True,
            skip_vars=self._skip_vars,
        )

        for image, array in zip(self._images, hexa_field.arrays):
            image.set_array(array.flatten())

        if self._QUIVER:
            hexa_vec_xaxis, hexa_vec_yaxis = phys_fields.get_vector_for_plot(
                time=time, skip_vars=self._skip_vars
            )

            vx_quiver, vy_quiver, vmax = set_of_phys_files.compute_vectors_quiver(
                hexa_vec_xaxis, hexa_vec_yaxis, self._indices_vectors_in_elems
            )

            if self.normalize_vectors:
                vx_quiver /= vmax
                vy_quiver /= vmax

            self._ani_quiver.set_UVC(vx_quiver, vy_quiver)
        else:
            vmax = None

        self.phys_fields._set_title(self.ax, self.key_field, time, vmax)

    def _set_clim(self):
        """Maintains a constant colorbar throughout the animation."""

        clim = self._clim
        if clim is not None:
            for image in self._images:
                image.set_clim(*clim)
            ticks = np.linspace(*clim, num=11, endpoint=True)
            self._ani_cbar.set_ticks(ticks)

    def _init_labels(self, xlabel=None, ylabel=None):
        """Initialize the labels."""

        (
            letter_x_axis,
            letter_y_axis,
        ) = self.phys_fields.set_of_phys_files.get_letters_axes_from_equation(
            self.phys_fields._equation
        )

        super()._init_labels(f"${letter_x_axis}$", f"${letter_y_axis}$")



"""Restart a fluidsim simulation (:mod:`fluidsim.util.scripts.restart`)

.. autoclass:: RestarterABC
   :members:
   :private-members:
   :undoc-members:

"""

from abc import ABCMeta, abstractmethod
import argparse
import sys
from pathlib import Path
from time import sleep
from textwrap import dedent

# import potentially needed for the exec
from math import pi

from fluiddyn.util import mpi

from fluidsim_core.scripts import parse_args


class RestarterABC(metaclass=ABCMeta):
    def create_parser(self):
        """Create the argument parser with default arguments"""
        parser = argparse.ArgumentParser(
            description="Restart a fluidsim simulation",
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

        parser.add_argument(
            "path",
            type=str,
            help="Path of the directory or file from which to restart",
        )
        parser.add_argument(
            "-oc",
            "--only-check",
            action="store_true",
            help="Only check what should be done",
        )
        parser.add_argument(
            "-oi",
            "--only-init",
            action="store_true",
            help="Only run initialization phase",
        )
        parser.add_argument(
            "--modify-params",
            type=str,
            default=None,
            help="Code modifying the `params` object.",
        )
        parser.add_argument(
            "--new-dir-results",
            action="store_true",
            help="Create a new directory for the new simulation",
        )
        return parser

    def restart(self, args=None, **defaults):
        """Restart a simulation"""
        parser = self.create_parser()

        if defaults:
            parser.set_defaults(**defaults)

        args = parse_args(parser, args)

        params, Simul = self._get_params_simul_class(args)

        params.NEW_DIR_RESULTS = args.new_dir_results

        self._set_params_time_stepping(params, args)

        if args.only_check or args.only_init:
            params.output.HAS_TO_SAVE = False

        if hasattr(args, "max_elapsed") and args.max_elapsed is not None:
            params.time_stepping.max_elapsed = args.max_elapsed

        path_restart_file = self._get_path_restart_file(params, args)

        if args.modify_params is not None:
            exec(args.modify_params)

        if args.only_check:
            mpi.printby0(params)
            return params, None

        try:
            self._check_params_time_stepping(params, path_restart_file, args)
        except ValueError:
            return params, None

        sim = Simul(params)

        if args.only_init:
            return params, sim

        self._start_sim(sim, args)

        mpi.printby0(
            self._str_command_after_simul.format(path_run=sim.output.path_run)
        )

        return params, sim

    _str_command_after_simul = dedent(
        """
        # To visualize with IPython:

        cd {path_run}; fluidsim-ipy-load
    """
    )

    @abstractmethod
    def _get_params_simul_class(self, args):
        "Create params object and Simul class"

    @abstractmethod
    def _set_params_time_stepping(self, params, args):
        "Set time stepping parameters"

    @abstractmethod
    def _start_sim(self, sim, args):
        "Start the new simulation"

    @abstractmethod
    def _check_params_time_stepping(self, params, path_file, args):
        "Check time stepping parameters"

    def _get_path_restart_file(self, params, args):
        path_file = Path(params.init_fields.from_file.path)
        mpi.printby0(path_file)
        return path_file

    def main(self):
        """Entry point fluidsim-restart"""
        params, sim = self.restart()

        if sim is not None and sim.time_stepping._has_to_stop:
            if mpi.rank == 0:
                # processes with rank 0 exits early with exit code 99 or 0
                exit_code = 99
                if (
                    Path(sim.output.path_run) / "IDEMPOTENT_NO_RELAUNCH"
                ).exists():
                    exit_code = 0
                print("Simulation is not completed and could be relaunched")

            if mpi.nb_proc > 1:
                mpi.comm.barrier()

            if mpi.rank == 0:
                sys.exit(exit_code)

            sleep(0.5)
            sys.exit()
